'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.parseRefs = exports.resolveVersion = exports.isCommitSha = undefined;

var _asyncToGenerator2;

function _load_asyncToGenerator() {
  return _asyncToGenerator2 = _interopRequireDefault(require('babel-runtime/helpers/asyncToGenerator'));
}

var _slicedToArray2;

function _load_slicedToArray() {
  return _slicedToArray2 = _interopRequireDefault(require('babel-runtime/helpers/slicedToArray'));
}

var _misc;

function _load_misc() {
  return _misc = require('../misc.js');
}

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const semver = require('semver');

const REF_TAG_PREFIX = 'refs/tags/';
const REF_BRANCH_PREFIX = 'refs/heads/';
const REF_PR_PREFIX = 'refs/pull/';

// This regex is designed to match output from git of the style:
//   ebeb6eafceb61dd08441ffe086c77eb472842494  refs/tags/v0.21.0
// and extract the hash and ref name as capture groups
const GIT_REF_LINE_REGEXP = /^([a-fA-F0-9]+)\s+(refs\/(?:tags|heads|pull)\/.*)$/;

const COMMIT_SHA_REGEXP = /^[a-f0-9]{5,40}$/;
const REF_NAME_REGEXP = /^refs\/(tags|heads)\/(.+)$/;

const isCommitSha = exports.isCommitSha = target => COMMIT_SHA_REGEXP.test(target);

const tryVersionAsGitCommit = (_ref) => {
  let version = _ref.version,
      refs = _ref.refs,
      git = _ref.git;

  const lowercaseVersion = version.toLowerCase();
  if (!isCommitSha(lowercaseVersion)) {
    return Promise.resolve(null);
  }
  for (const _ref2 of refs.entries()) {
    var _ref3 = (0, (_slicedToArray2 || _load_slicedToArray()).default)(_ref2, 2);

    const ref = _ref3[0];
    const sha = _ref3[1];

    if (sha.startsWith(lowercaseVersion)) {
      return Promise.resolve({ sha, ref });
    }
  }
  return git.resolveCommit(lowercaseVersion);
};

const tryEmptyVersionAsDefaultBranch = (_ref4) => {
  let version = _ref4.version,
      git = _ref4.git;
  return version.trim() === '' ? git.resolveDefaultBranch() : Promise.resolve(null);
};

const tryWildcardVersionAsDefaultBranch = (_ref5) => {
  let version = _ref5.version,
      git = _ref5.git;
  return version === '*' ? git.resolveDefaultBranch() : Promise.resolve(null);
};

const tryRef = (refs, ref) => {
  const sha = refs.get(ref);
  return sha ? { sha, ref } : null;
};

const tryVersionAsFullRef = (_ref6) => {
  let version = _ref6.version,
      refs = _ref6.refs;
  return version.startsWith('refs/') ? tryRef(refs, version) : null;
};

const tryVersionAsTagName = (_ref7) => {
  let version = _ref7.version,
      refs = _ref7.refs;
  return tryRef(refs, `${REF_TAG_PREFIX}${version}`);
};

const tryVersionAsPullRequestNo = (_ref8) => {
  let version = _ref8.version,
      refs = _ref8.refs;
  return tryRef(refs, `${REF_PR_PREFIX}${version}`);
};

const tryVersionAsBranchName = (_ref9) => {
  let version = _ref9.version,
      refs = _ref9.refs;
  return tryRef(refs, `${REF_BRANCH_PREFIX}${version}`);
};

const computeSemverNames = (_ref10) => {
  let config = _ref10.config,
      refs = _ref10.refs;

  const names = {
    tags: [],
    heads: []
  };
  for (const ref of refs.keys()) {
    const match = REF_NAME_REGEXP.exec(ref);
    if (!match) {
      continue;
    }

    var _match = (0, (_slicedToArray2 || _load_slicedToArray()).default)(match, 3);

    const type = _match[1],
          name = _match[2];

    if (semver.valid(name, config.looseSemver)) {
      names[type].push(name);
    }
  }
  return names;
};

const findSemver = (version, config, namesList) => config.resolveConstraints(namesList, version);

const tryVersionAsTagSemver = (() => {
  var _ref11 = (0, (_asyncToGenerator2 || _load_asyncToGenerator()).default)(function* (_ref12, names) {
    let version = _ref12.version,
        config = _ref12.config,
        refs = _ref12.refs;

    const result = yield findSemver(version, config, names.tags);
    return result ? tryRef(refs, `${REF_TAG_PREFIX}${result}`) : null;
  });

  return function tryVersionAsTagSemver(_x, _x2) {
    return _ref11.apply(this, arguments);
  };
})();

const tryVersionAsBranchSemver = (() => {
  var _ref13 = (0, (_asyncToGenerator2 || _load_asyncToGenerator()).default)(function* (_ref14, names) {
    let version = _ref14.version,
        config = _ref14.config,
        refs = _ref14.refs;

    const result = yield findSemver(version, config, names.heads);
    return result ? tryRef(refs, `${REF_BRANCH_PREFIX}${result}`) : null;
  });

  return function tryVersionAsBranchSemver(_x3, _x4) {
    return _ref13.apply(this, arguments);
  };
})();

const tryVersionAsSemverRange = (() => {
  var _ref15 = (0, (_asyncToGenerator2 || _load_asyncToGenerator()).default)(function* (options) {
    const names = computeSemverNames(options);
    return (yield tryVersionAsTagSemver(options, names)) || tryVersionAsBranchSemver(options, names);
  });

  return function tryVersionAsSemverRange(_x5) {
    return _ref15.apply(this, arguments);
  };
})();

const VERSION_RESOLUTION_STEPS = [tryEmptyVersionAsDefaultBranch, tryVersionAsGitCommit, tryVersionAsFullRef, tryVersionAsTagName, tryVersionAsPullRequestNo, tryVersionAsBranchName, tryVersionAsSemverRange, tryWildcardVersionAsDefaultBranch];

/**
 * Resolve a git-url hash (version) to a git commit sha and branch/tag ref
 * Returns null if the version cannot be resolved to any commit
 */

const resolveVersion = exports.resolveVersion = (() => {
  var _ref16 = (0, (_asyncToGenerator2 || _load_asyncToGenerator()).default)(function* (options) {
    for (const testFunction of VERSION_RESOLUTION_STEPS) {
      const result = yield testFunction(options);
      if (result !== null) {
        return result;
      }
    }
    return null;
  });

  return function resolveVersion(_x6) {
    return _ref16.apply(this, arguments);
  };
})();

/**
 * Parse Git ref lines into hash of ref names to SHA hashes
 */

const parseRefs = exports.parseRefs = stdout => {
  // store references
  const refs = new Map();

  // line delimited
  const refLines = stdout.split('\n');

  for (const line of refLines) {
    const match = GIT_REF_LINE_REGEXP.exec(line);

    if (match) {
      var _match2 = (0, (_slicedToArray2 || _load_slicedToArray()).default)(match, 3);

      const sha = _match2[1],
            tagName = _match2[2];

      // As documented in gitrevisions:
      //   https://www.kernel.org/pub/software/scm/git/docs/gitrevisions.html#_specifying_revisions
      // "A suffix ^ followed by an empty brace pair means the object could be a tag,
      //   and dereference the tag recursively until a non-tag object is found."
      // In other words, the hash without ^{} is the hash of the tag,
      //   and the hash with ^{} is the hash of the commit at which the tag was made.

      const name = (0, (_misc || _load_misc()).removeSuffix)(tagName, '^{}');

      refs.set(name, sha);
    }
  }

  return refs;
};